#include "ossmac.h"

#include <stdio.h>
#include <assert.h>

#include <Types.h>
#include <QuickDraw.h>
#include <Files.h>
#include <MacWindows.h>
#include <Dialogs.h>
#include <Events.h>
#include <Controls.h>
#include <LowMem.h>
#include <Processes.h>
#include <AppleEvents.h>
#include <Errors.h>

#include "GetFolder.h"


OSType		gAppCreator = '????';
short		gAppResFile = 0;


// =======================================================================================
// == State-saving object methods.
// ==

#ifdef __cplusplus

TFontState::TFontState()
{
	#ifndef NDEBUG
	GetPort(&fPort);
	#endif
	fFont = qd.thePort->txFont; fSize = qd.thePort->txSize;
	fFace = qd.thePort->txFace; fMode = qd.thePort->txMode;
}

TFontState::~TFontState()
{
	assert(fPort == qd.thePort);
	if (fFont != qd.thePort->txFont) TextFont(fFont);
	if (fSize != qd.thePort->txSize) TextSize(fSize);
	if (fFace != qd.thePort->txFace) TextFace(fFace);
	if (fMode != qd.thePort->txMode) TextMode(fMode);
}

TFractEnableState::TFractEnableState()
	{ fEnable = LMGetFractEnable(); }

TFractEnableState::TFractEnableState(UInt8 enable)
	{ fEnable = LMGetFractEnable(); SetFractEnable(enable); }

TFractEnableState::~TFractEnableState()
	{ SetFractEnable(fEnable); }

#endif

// =======================================================================================
// == Debugging routines
// ==

#if DEBUG_IO

#define TRACK_FILE_NAMES_USING_STL 0

#if TRACK_FILE_NAMES_USING_STL
	#include <map.h>
#endif
#include "osmac.h"

#if TRACK_FILE_NAMES_USING_STL
typedef map<osfildef *, const_string, less<osfildef *>, allocator<const_string> > filename_map;
	/* Last two template arguments are the default. */
#else
class filename_map {
	public:
		struct value_type {
			osfildef *first;
			char      second[256];
			
			value_type()
				{ first = NULL; second[0] = '\0'; }
			value_type(osfildef *fp, const char *fname)
				{ first = fp; strcpy(second, fname); }
			value_type(const value_type& x)
				{ first = x.first; strcpy(second, x.second); }
			value_type& operator=(const value_type& x)
				{ first = x.first; strcpy(second, x.second); return *this; }
		};
		
		typedef osfildef*  key_type;
		typedef value_type*  iterator;
		
	private:
		value_type	items[16];
	
	public:
		iterator begin()  { return &items[0]; }
		iterator end()    { return &items[16]; }
		
		iterator insert(const value_type& x);
		iterator find(const key_type& k);
		void erase(iterator x);
};


filename_map::iterator filename_map::insert(const filename_map::value_type& x)
{
	for (iterator i = begin(); i != end(); i++) {
		if ((*i).first == NULL) {
			(*i) = x;
			return i;
		}
	}
	printf("[Couldnt find empty space!]");
	return NULL;
}

filename_map::iterator filename_map::find(const filename_map::key_type& k)
{
	iterator i;
	
	for (i = begin(); i != end(); i++)
		if ((*i).first == k) break;
	
	return i;
}

void filename_map::erase(filename_map::iterator x)
{
	if (x >= begin() && x < end())
		(*x) = value_type();
}
#endif

static filename_map gsFileMap;


void track_fileopen(const char *routine, const char *fname, osfildef *fp)
{
	if (fp) {
		printf("[F: %s(%s) --> %x]\n", routine, fname, fp);
		gsFileMap.insert(filename_map::value_type(fp, fname));
	}
	else {
		printf("[F: %s(%s) failed.]\n", routine, fname);
	}
}

void track_fileclose(const char *routine, osfildef *fp)
{
	if (fp) {
		#if 1
		filename_map::iterator  f;
		
		f = gsFileMap.find(fp);
		if (f == gsFileMap.end()) {
			printf("[F: %s(%x) -- couldnt find %x in list!]\n", routine, fp, fp);
		}
		else {
			printf("[F: %s(%x \"%s\")]\n", routine, fp, (const char *) (*f).second);
			gsFileMap.erase(f);
		}
		#else
		printf("[F: %s(%x)]\n", routine, fp);
		#endif
	}
	else {
		printf("[F: %s(NULL) failed.]\n", routine);
	}
}

#endif


// =======================================================================================
// == Toolbox initialization
// ==

#define FAIL_ERR(err, x)  \
	if ((err = (x)) != noErr) \
		{ DPRINTF((#x " failed with error %d.\n", err)); goto error; } \
	else \
		((void) 0)

OSErr InstallStandardAEHandlers();
void RemoveStandardAEHandlers();

OSErr InitToolbox(void)
{
	OSErr	err;
	
	#if DEBUG
	static bool	done_init = false;
	
	assert(!done_init);
	done_init = true;
	#endif
	
	SetApplLimit(GetApplLimit() - (48L * 1024L));
	MaxApplZone();
	
	InitGraf(&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(NULL);
	InitCursor();
	FlushEvents(everyEvent, 0);

	gAppResFile = CurResFile();
	
	err = InstallStandardAEHandlers();
	
	return err;
}


OSErr GetApplicationInfo(StringPtr name, FSSpec *fsspec)
{
	ProcessSerialNumber	psn;
	ProcessInfoRec		pi;
	OSErr				err;
	
	FAIL_ERR(err, GetCurrentProcess(&psn));
	pi.processInfoLength = sizeof(pi);
	pi.processName = name;
	pi.processAppSpec = fsspec;
	FAIL_ERR(err, GetProcessInformation(&psn, &pi));
	gAppCreator = pi.processSignature;
error:
	return err;
}


static AEEventHandlerUPP gsOpenApp = NULL;
static AEEventHandlerUPP gsOpenDoc = NULL;
static AEEventHandlerUPP gsPrintDoc = NULL;
static AEEventHandlerUPP gsQuitApp = NULL;

// We use OSErr instead of OSStatus because this can be returned by an AppleEvent
static OSErr CheckAppleEventForMissingParams(AppleEvent *event)
{
	OSErr		err;
	Size 		actualSize;
	DescType	returnedType;

	err = AEGetAttributePtr(event, keyMissedKeywordAttr, typeWildCard, &returnedType, NULL, 0, &actualSize);
	
	// If we found a missing parameter, we should ignore this event.
	if (err == noErr)
		err = errAEEventNotHandled;
	// Otherwise, if we couldn't find a missing attribute, the event should be handled.
	else if (err == errAEDescNotFound)
		err = noErr;	
	
	return err;
}

static pascal OSErr HandleOpenAppEvent(AppleEvent *event, AppleEvent *reply, long refCon)
{
	#pragma unused(reply, refCon)
	
	OSErr 	err;
	
	if (VERIFY((err = CheckAppleEventForMissingParams(event)) == noErr))
		err = HandleOpenApp();
	
	return err;
}

static pascal OSErr HandleOpenDocumentEvent(AppleEvent *event, AppleEvent *reply, long refCon)
{
	#pragma unused(reply, refCon)
	
	OSErr 			err;
	AEDescList		docsList;
	SInt32			docCount;
	
	docsList.dataHandle = NULL;
	FAIL_ERR(err, AEGetParamDesc(event, keyDirectObject, typeAEList, &docsList));
	FAIL_ERR(err, CheckAppleEventForMissingParams(event));
	
	// Find out how many items were passed into us.
	FAIL_ERR(err, AECountItems(&docsList, &docCount));
	
	for (SInt32 i = 1; i <= docCount; i++) {
		FSSpec		file;
		AEKeyword	keyword;
		DescType	type;
		Size		docSize;
		
		if (VERIFY(AEGetNthPtr(&docsList, i, typeFSS, &keyword, &type,
								&file, sizeof(file), &docSize) == noErr))
			(void) HandleOpenDocument(&file);
	}
	
error:
	if (docsList.dataHandle != NULL)
		(void) AEDisposeDesc(&docsList);
	return err;
}

static pascal OSErr HandlePrintDocumentEvent(AppleEvent *event, AppleEvent *reply, long refCon)
{
	#pragma unused(event, reply, refCon)
	
	// We don't do printing.
	return errAEEventNotHandled;
}

static pascal OSErr HandleQuitAppEvent(AppleEvent *event, AppleEvent *reply, long refCon)
{
	#pragma unused(reply, refCon)
	
	OSErr	err;
	
	if (VERIFY((err = CheckAppleEventForMissingParams(event)) == noErr))
		HandleQuitApp();
	
	return err;
}

OSErr InstallStandardAEHandlers()
{
	OSErr	err = noErr;
	
	// Build the UPPs for our standard event handlers
	gsOpenApp  = NewAEEventHandlerProc(HandleOpenAppEvent);
	gsOpenDoc  = NewAEEventHandlerProc(HandleOpenDocumentEvent);
	gsPrintDoc = NewAEEventHandlerProc(HandlePrintDocumentEvent);
	gsQuitApp  = NewAEEventHandlerProc(HandleQuitAppEvent);

	// Install the event handlers into the Application's Apple Event table.
	FAIL_ERR(err, AEInstallEventHandler(kCoreEventClass, kAEOpenApplication, gsOpenApp, 0, false));
	FAIL_ERR(err, AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments, gsOpenDoc, 0, false));
	FAIL_ERR(err, AEInstallEventHandler(kCoreEventClass, kAEPrintDocuments, gsPrintDoc, 0, false));
	FAIL_ERR(err, AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, gsQuitApp, 0, false));

error:
	return err;
}


static OSErr RemoveCoreAEHandler(AEEventHandlerUPP *upp)
{
	if (*upp) {
		OSErr	err;
		
		err = AERemoveEventHandler(kCoreEventClass, kAEOpenApplication, *upp, false);
		DisposeRoutineDescriptor(*upp);
		*upp = NULL;
		return err;
	}
	else
		return noErr;
}

void RemoveStandardAEHandlers()
{
	(void) RemoveCoreAEHandler(&gsOpenApp);
	(void) RemoveCoreAEHandler(&gsOpenDoc);
	(void) RemoveCoreAEHandler(&gsPrintDoc);
	(void) RemoveCoreAEHandler(&gsQuitApp);
}
	
void HandleAllHighLevelEvents()
{
	EventRecord		ev;
	
	while (GetNextEvent(highLevelEventMask, &ev)) {
		OSErr		err;
		
		err = AEProcessAppleEvent(&ev);
		assert(err == noErr);
	}
}

#undef FAIL_ERR

// =======================================================================================
// == Dialog routines
// ==

#ifndef __APPEARANCE__
pascal OSErr GetDialogItemAsControl(DialogPtr inDialog, SInt16 inItemNo, ControlHandle *outControl)
{
	assert(inDialog != NULL);
	if (inDialog) {
		short	itemType;
		Handle	item;
		Rect	itemRect;
		
		GetDialogItem(inDialog, inItemNo, &itemType, &item, &itemRect);
		if (itemType & kControlDialogItem) {
			*outControl = (ControlHandle) item;
			return noErr;
		}
		else {
			DPRINTF(("GetDialogItemAsControl called on non-control item: inDialog %x, item %d", inDialog, inItemNo));
			*outControl = NULL;
			return 1;
		}
	}
	else {
		*outControl = NULL;
		return 2;
	}
}
#endif

SInt16 GetControlItemValue(DialogPtr dialog, short item)
{
	ControlHandle	control;
	
	if (GetDialogItemAsControl(dialog, item, &control) == noErr)
		return GetControlValue(control);
	else
		return 0;
}

void SetControlItemValue(DialogPtr dialog, short item, SInt16 newvalue)
{
	ControlHandle	control;
	
	if (GetDialogItemAsControl(dialog, item, &control) == noErr)
		SetControlValue(control, newvalue);
}


#define IsTextItem(it) \
	((((it) & ~kItemDisableBit) == kStaticTextDialogItem) || \
	 (((it) & ~kItemDisableBit) == kEditTextDialogItem))

void GetControlItemTitle(DialogPtr dialog, short itemNo, Str255 title)
{
	#ifdef __APPEARANCE__
	ControlHandle	control;
	
	title[0] = '\0';
	if (GetDialogItemAsControl(dialog, itemNo, &control) == noErr)
		GetControlTitle(control, title);
	#else
	/*
	 ## We engage in a bit of fakery here to pretend that static text
	 ## and edit text items are controls when we're not using the
	 ## Appearance Manager.
	*/
	title[0] = '\0';
	if (dialog) {
		short	itemType;
		Handle	item;
		Rect	itemRect;
		
		GetDialogItem(dialog, itemNo, &itemType, &item, &itemRect);
		if (itemType & kControlDialogItem)
			GetControlTitle((ControlHandle) item, title);
		else if (IsTextItem(itemType))
			GetDialogItemText(item, title);
		else {
			DPRINTF(("GetControlItemTitle called on unrecognized item type!  dialog %x, item %d", dialog, itemNo));
		}
	}
	#endif
}

void SetControlItemTitle(DialogPtr dialog, short itemNo, ConstStr255Param title)
{
	#ifdef __APPEARANCE__
	ControlHandle	control;
	
	if (GetDialogItemAsControl(dialog, itemNo, &control) == noErr)
		SetControlTitle(control, title);
	#else
	/*
	 ## We engage in a bit of fakery here to pretend that static text
	 ## and edit text items are controls when we're not using the
	 ## Appearance Manager.
	*/
	if (dialog) {
		short	itemType;
		Handle	item;
		Rect	itemRect;
		
		GetDialogItem(dialog, itemNo, &itemType, &item, &itemRect);
		if (itemType & kControlDialogItem)
			SetControlTitle((ControlHandle) item, title);
		else if (IsTextItem(itemType))
			SetDialogItemText(item, title);
		else {
			DPRINTF(("SetControlItemTitle called on unrecognized item type!  dialog %x, item %d", dialog, itemNo));
		}
	}
	#endif
}

void GetControlItemRect(DialogPtr dialog, short itemNo, Rect *bounds)
{
	#ifdef __APPEARANCE__
	ControlHandle	control;
	
	if (GetDialogItemAsControl(dialog, itemNo, &control) == noErr)
		*bounds = (*control)->contrlRect;
	else
		SetRect(bounds, 0, 0, 0, 0);
	
	#if DEBUG
	{
		Rect		itemRect;
		
		if (dialog) {
			short	itemType;
			Handle	item;
			
			GetDialogItem(dialog, itemNo, &itemType, &item, itemRect);
		}
		else
			SetRect(&itemRect, 0, 0, 0, 0);
		
		assert(EqualRect(&itemRect, bounds));
	}
	#endif
	
	#else
	if (dialog) {
		short	itemType;
		Handle	item;
		
		GetDialogItem(dialog, itemNo, &itemType, &item, bounds);
	}
	else
		SetRect(bounds, 0, 0, 0, 0);
	#endif
}

void HiliteControlItem(DialogPtr dialog, short itemNo, ControlPartCode hiliteState)
{
	#ifdef __APPEARANCE__
	ControlHandle	control;
	
	if (GetDialogItemAsControl(dialog, itemNo, &control) == noErr)
		HiliteControl(control, hiliteState);
	#else
	/*
	 ## This long-hand implementation is to keep our local implementation of
	 ## GetDialogItemAsControl from yakking if it sees a static or edit text
	 ## item, which Appearance Manager could deal with, but we can't.
	*/
	if (dialog) {
		short	itemType;
		Handle	item;
		Rect	itemRect;
		
		GetDialogItem(dialog, itemNo, &itemType, &item, &itemRect);
		if (itemType & kControlDialogItem)
			HiliteControl((ControlHandle) item, hiliteState);
		else if (itemType == kStaticTextDialogItem || itemType == kEditTextDialogItem)
			;
		else {
			DPRINTF(("HiliteControlItem called on non-control!  dialog %x, item %d", dialog, itemNo));
		}
	}
	#endif
}

bool ToggleControlItemValue(DialogPtr dialog, short itemNo)
{
	SInt16	value;
	
	value = GetControlItemValue(dialog, itemNo);
	SetControlItemValue(dialog, itemNo, !value);
	return !value;
}							

void drawbox(DialogPtr dialog, short itemNo)
{
	Rect	itemRect;

	GetControlItemRect(dialog, itemNo, &itemRect);
	InsetRect(&itemRect, -1, -1);
	FrameRect(&itemRect);
}


// =======================================================================================
// == File routines
// ==

/*
 *	DirToPath, VRefToPath, FSSpecToPath:
 *
 *  Service routines to convert Mac file and directory specifications to full
 *	pathnames, which are required by the C-style I/O routines used by most of
 *	TADS.  The mixed-case versions give a Pascal string; the lowercase versions
 *	give a C string.
 */
OSErr DirToPath(short vRefNum, long dirID, Str255 s)
{
	CInfoPBRec	block;
	Str255		dirname;
	OSErr		err = noErr;
	
	s[0] = '\0';
	block.dirInfo.ioNamePtr = dirname;
	block.dirInfo.ioDrParID = dirID;
	do {
		block.dirInfo.ioVRefNum = vRefNum;
		block.dirInfo.ioFDirIndex = -1;
		block.dirInfo.ioDrDirID = block.dirInfo.ioDrParID;
		
		err = PBGetCatInfo(&block, false);
		
		dirname[++dirname[0]] = ':';
		StringCat_pp(dirname, s, sizeof(dirname));
		StringCopy_pp(s, dirname, sizeof(Str255));
	} while (block.dirInfo.ioDrDirID != fsRtDirID && err == noErr);
	
	return err;
}

OSErr dirtopath(short vRefNum, long dirID, char *s)
{
	OSErr	err;
	Str255	path;
	
	err = DirToPath(vRefNum, dirID, path);
	StringCopy_cp(s, path, 0);
	return err;
}


OSErr VRefToPath(short vRef, Str255 s)
{
	long	dirID;
	long	procID;
	short	vRefNum;
	OSErr	err;
		
	err = GetWDInfo(vRef, &vRefNum, &dirID, &procID);
	return (err == noErr) ? DirToPath(vRefNum, dirID, s) : err;
}

OSErr vreftopath(short vRef, char *s)
{
	long	dirID;
	long	procID;
	short	vRefNum;
	OSErr	err;
		
	err = GetWDInfo(vRef, &vRefNum, &dirID, &procID);
	return (err == noErr) ? dirtopath(vRefNum, dirID, s) : err;
}


OSErr FSSpecToPath(const FSSpec *fsspec, Str255 s)
{
	OSErr	err;
	
	if ((err = DirToPath(fsspec->vRefNum, fsspec->parID, s)) == noErr)
		StringCat_pp(s, fsspec->name, sizeof(Str255));
	
	return err;
}

OSErr fsspectopath(const FSSpec *fsspec, char *s)
{
	OSErr	err;
	
	if ((err = dirtopath(fsspec->vRefNum, fsspec->parID, s)) == noErr)
		StringCat_cp(s, fsspec->name, 256);
	
	return err;
}


static void os0askfile__internal(const char *prompt, SFReply *reply, OSType type, bool writing)
{
	SFTypeList		myTypes;
	Point			sfgWhere = { 90, 82 };
	Point			sfpWhere = { 106, 104 };
	Str255			pascalprompt;
	static Str255	lastsave = "\pUntitled";
	
	StringCopy_pc(pascalprompt, prompt, sizeof(pascalprompt));
	myTypes[0] = type;
	
	if (writing)
		SFPutFile(sfpWhere, pascalprompt, lastsave, 0L, reply);
	else
		SFGetFile(sfgWhere, pascalprompt, 0L, 1, myTypes, 0L, reply);
}

OSErr os0askfile(const char *prompt, char *fnp, size_t fnl, OSType type, bool wrtflag)
{
	SFReply			reply;
	OSErr			err;

	os0askfile__internal(prompt, &reply, type, wrtflag);
	if (reply.good) {
		if ((err = vreftopath(reply.vRefNum, fnp)) != noErr)
			return err;
		StringCat_cp(fnp, reply.fName, fnl);
		return noErr;
	}
	else
		return 1;
}

OSErr os0askfile(const char *prompt, Str255 fname, OSType type, bool wrtflag)
{
	SFReply			reply;
	OSErr			err;

	os0askfile__internal(prompt, &reply, type, wrtflag);
	if (reply.good) {
		if ((err = VRefToPath(reply.vRefNum, fname)) != noErr)
			return err;
		StringCat_pp(fname, reply.fName, sizeof(Str255));
		return noErr;
	}
	else
		return 1;
}


OSErr osaskdir(char *fnp, int fnl)
{
	/* XXX: This doesn't have any option to insist on a writable directory.
		For now, this isn't a problem, since it's only used to get source
		folders, but it would be a good thing to support. */
	
	StandardFileReply	reply;
	
	StandardGetFolder("\pSelect Folder", false, false, &reply);
	if (reply.sfGood) {
		OSErr	err;
		
		err = fsspectopath(&reply.sfFile, fnp);
		
		// Add the trailing ':', because fsspectopath doesn't.	
		if (err == noErr) {
			size_t i = strlen(fnp);
			
			if (i < fnl - 1) {
				fnp[i] = ':';
				fnp[i+1] = '\0';
			}
		}
		
		return err;
	}
	else
		return 1;
}


// =======================================================================================
// == Other.
// ==

void XXX_FixStandAlone(void)
/*
 * This function fixes a mysterious problem in _RTExit which
 * causes the app to crash.
 */
{
	#if MANGLE_ISSTANDALONE
	#if !__powerc
	/* This problem is specific to the 68K build. */
	extern int StandAlone, _IsStandAlone;

	if (_IsStandAlone != StandAlone)
		_IsStandAlone = StandAlone;
	else
		DPRINTF(("Hey! StandAlone == _IsStandAlone!  (This is good, hit cmd-g to continue.)"));
	#endif
	#endif
}


